; File: synmidi.asm
; Code for Synergy MIDI interface 	R. Grieb
; 9/19/17 Initial work, started with MIDI.asm
; Use tab setting of 8 if using PFE to look at this file
; (under current Modes, Text formatting)
; 10/1/17 Changed register for TXIF to PIR1 instead of PIE1
; 10/2/17 Changed to using FIFOcnt to track FIFO contents
;         Changed FULL to come on at 132 bytes, (not 133!)
; 10/7/17 Receive seemed to be working fine.   TX not tested.
;	  Adding code to change banks now.
; 	  There is some sort of MIDI rx issue, with data being 
;	  corrupted and dropped. 
;	  Changing things around a bit. New version seems solid.
;	  More testing needed, of course, but looks good. 

; To do:

        LIST    P=18F2420	;PIC18F2420 is the target processor
			
			;include std definitions for our part
#include "P18F2420.INC"

	CONFIG OSC=HSPLL, FCMEN=OFF, IESO=OFF
	CONFIG PWRT=ON, BOREN=SBORDIS, BORV=2
	CONFIG WDT=OFF, WDTPS=32768
	CONFIG CCP2MX=PORTC, LPT1OSC=ON, PBADEN=OFF, MCLRE=OFF
	CONFIG STVREN=OFF, LVP=OFF, XINST=OFF
	CONFIG DEBUG=OFF
	CONFIG CP0=ON, CP1=ON
	CONFIG CPD=OFF, CPB=ON
	CONFIG WRT0=OFF, WRT1=OFF
	CONFIG WRTC=OFF, WRTB=OFF, WRTD=OFF
	CONFIG EBTR0=OFF, EBTR1=OFF, EBTRB=OFF

; resulting config words:
;		300001	08
;		300002	0A
;		300003	1E
;		300005	05
;		300006	80
;		300008	00
;		300009	80
;		30000A	03
;		30000B	E0
;		30000C	03
;		30000D	40

;#define simtest 1
;#define dbgenab 1

;Our I/O port definitions:

#define IBNK0	PORTA,0		;output, cartridge selection bank
#define IBNK1	PORTA,1		;output
#define IBNK2	PORTA,2		;output
#define IBNK3	PORTA,3		;output
#define IBNK4	PORTA,4		;output
#define FLIPTBE	PORTA,5		;output
#define SPARE1	PORTA,6		;input, osc will override
#define SPARE2	PORTA,7		;output, osc will override

#define PORTADIR 0x40		;pin A6 is input

#define TBE	PORTB,0		;input, transmit buffer empty
#define DRDY	PORTB,1		;input, rec'd data ready
#define CLKRDAT	PORTB,2		;output
#define RDWDATb	PORTB,3		;output
#define MIDID4	PORTB,4		;output at startup
#define MIDID5	PORTB,5		;output
#define MIDID6	PORTB,6		;output
#define MIDID7	PORTB,7		;output

#define PORTBDIR 0x03		;B port 0 and 1 are inputs

#define MIDID0	PORTC,0		;output at startup
#define MIDID1	PORTC,1		;output
#define MIDID2	PORTC,2		;output
#define MIDID3	PORTC,3		;output
#define FULL	PORTC,4		;output, FIFO flag bit
#define EMPTY	PORTC,5		;output, FIFO flag bit
#define SERTX	PORTC,6		;output, serial data output
#define SERRX	PORTC,7		;input, MIDI serial data input

#define PORTCDIR 0x80		;pin C7 is input
				
				;used by:
;FSR0				;
;FSR1				;buffer read ptr
;FSR2				;buffer write ptr

;Our variables:
				;used by:
temp1		equ	02	;delay routines,cnvrtcf
temp2		equ	03	;delay routines,cnvrtcf
temp3		equ	04	;delay routines,copytab,dspudpar,dspdPC,dspmem,sndbhexsp
temp4		equ	05	;cnvrtcf,copytab,dspbhex1,sndbhex1,getind,getinput,regpre,sendbhex
temp5		equ	06	;cnvrtcf,doWcmd,dspnn,wrmem,mdtest,measclk
itemp1		equ	07	;myisr	
flags1		equ	08	
flags2		equ	09	;settings that will be saved in EEPROM
ASCtemp		equ	0A
loopctr		equ	0B	;softres,clrbp,cnvrtcf,listmem,listop
rxdata		equ	0C	;serial port rec'd data
MIDIbyte	equ	0D
portbval	equ	0E	
portcval	equ	0F

				;upper byte of MIDI buffer address is always 0x02
;MIDIwrptr	equ	20	;pts to last location we wrote to
;MIDIrdptr	equ	21	;pts to last location we read
				;when the ptrs are equal, the buffer is empty
				;when wrptr = rdptr + 1, there is one byte in the buffer
				;when rdptr = wrptr + 1, the buffer is full
FIFOcnt		equ	22	;number of bytes in FIFO

testctr		equ	23	;for test simulation
txbyte		equ	24	;temporary storage for MIDI out byte
rxbyte		equ	25	;rec'd MIDI data

flshrdatL	equ	71
flshrdatH	equ	72
flshaddL	equ	73	
flshaddH	equ	74
		
W_TEMP		equ	7E	;in access bank
STATUS_TEMP	equ	7F	;ditto

#define NEWKYDN	flags1,0	;
#define NEWKYUP	flags1,1	;
#define OVRRUN	flags1,2	;set when we have a serial data RX overrun condition
#define GOTNOTE flags1,3
#define DOUT	flags1,4	;set when we write a byte to register for CPU to read
#define NEWRX	flags1,5	;set when we read a new byte from serin

#define LOWNOTE	flags2,0	;these are saved in EEPROM
#define HGHNOTE flags2,1
#define LSTNOTE flags2,2

MIDIbuf0	equ	0x200	;MIDI serial receive buffer
				;but we limit contents to 132 bytes, to match FIFO
MIDIbufend	equ	0x2FF	;end address of buffer


				;ASCII characters
BELLCHAR	equ	0x07	;ring the bell
BSCHAR		equ	0x08	;backspace
LFCHAR		equ	0X0A	;line feed
RETCHAR		equ	0X0D	;carr ret
SPACECHAR	equ	0x20
TICK		equ	0x27
CTLC		equ	0x03	;control-C
CTLQ		equ	0x11	;control-Q
CTLS		equ	0x13	;control-S

				;debug LCD special characters
SETROW0		equ	'{'	;use chars hyperterm can make to ease testing
SETROW1		equ	'}'
SETROW2		equ	'['
SETROW3		equ	']'
CLRSCR		equ	'!'
CRCHAR		equ	0x0D
LFCHAR		equ	0x0A	

	org	0000h
	goto	start		

	org	0008h
	goto	myISR

				;LUT routines are out of order, to insure that
				;they don't cross a 256-byte boundary

getASC:	andlw	0x0F		;just use low nibble
	movwf	ASCtemp
	movf	PCL,W		;needed to load pclatH
	rlncf	ASCtemp,W	;as addwf doesn't seem to do it
	addwf	PCL,F		;returns ASCII for nibble in W
	retlw	A'0'	
	retlw	A'1'
	retlw	A'2'
	retlw	A'3'	
	retlw	A'4'
	retlw	A'5'
	retlw	A'6'	
	retlw	A'7'
	retlw	A'8'
	retlw	A'9'	
	retlw	A'A'
	retlw	A'B'
	retlw	A'C'	
	retlw	A'D'
	retlw	A'E'
	retlw	A'F'

			
				;PIC18 series code:
				;reads one 16-bit word of Flash into flshrdatl,h registers
flashrd:			;and bumps TBLPTR as well
	TBLRD*+			;read low Flash byte into latch, incr ptr
	movf	TABLAT,W
	movwf	flshrdatL	;save it
			
	TBLRD*+			;read high Flash byte into latch, incr ptr
	movf	TABLAT,W
	movwf	flshrdatH	;save it

	return



initser:			;sets up serial port for TX and receive
	movlw	0x08
	movwf	BAUDCON		;BRG16 = 1
	clrf	TXSTA		;BRGH = 0, SYNC = 0, tx not enabled
	clrf	SPBRGH
	movlw	D'51'		;select 31250 baud for MIDI
	movwf	SPBRG

	movlw	0xC0
	movwf	TRISC		;ser port needs both of these set	
	
	clrf	RCSTA
	nop	
	bsf	RCSTA,SPEN	;enable ser port, 8-bit, async, low speed
	nop
	bsf	RCSTA,CREN	;enable continuous reception
	movf	RCREG,W		;read data fifo
	movf	RCREG,W		;twice
	bsf	TXSTA,TXEN	;enable transmit	
	bcf	PIR1,RCIF	;clear flag
	bsf	PIE1,RCIE	;enable serial receive interrupt
	
	return



				;PIC18 series code:
ldtblptr:			;load table pointer used to access Flash array	
	movf 	flshaddL,W	;this is a byte address, probably needs to be word-aligned
	movwf 	TBLPTRL		;LS byte of program address to read
	movf 	flshaddH,W 
	movwf 	TBLPTRH		;MS byte of program address to read
	clrf	TBLPTRU		;bits above A15 all 0
	return	


				;our interrupt handler
				;use Microchip's code to save/restore status, W reg
myISR:	movwf	W_TEMP 		;copy W to TEMP register
	swapf	STATUS,W 	;swap status to be saved into W
	clrf	STATUS
	movwf 	STATUS_TEMP 	;save swapped status to STATUS_TEMP register

	btfss	PIR1,RCIF	;check rx data int flag
	bra	intdone

	movf	RCREG,W		;get byte, this will clear flag
	movwf	rxbyte		;save it
	bsf	NEWRX		;and set flag

intdone:
	swapf 	STATUS_TEMP,W 	;swap STATUS_TEMP register into W
				;(sets bank to original state)
	movwf 	STATUS 		;move W into STATUS register
	swapf 	W_TEMP,F 	;swap W_TEMP
	swapf 	W_TEMP,W 	;swap W_TEMP into W (swap doesn't set flags)

	retfie			;return from interrupt, also sets GIE



				;read user settings from EEPROM
rdcfg:	clrf	EEADR		;starting EEPROM address is 0

	bcf	EECON1,EEPGD	;select data memory
	bcf	EECON1,CFGS	;access EEPROM
	bsf	EECON1,RD	;kick off read
	movf	EEDATA,W	;read data
	movwf	flags2		;save it

	incf	EEADR,F		;complemented data
	bsf	EECON1,RD	;kick off read
	comf	EEDATA,W	;read data, invert it
	subwf	flags2,W	;subtract from first byte
	bz	donerc
	clrf	flags2		;put flags in default state
	bsf	LOWNOTE
	call	wrcfg		;write out default values, since EEPROM was not set up		
donerc:	return


	ifdef	dbgenab

sndCRLF:			;send carriage return-line feed
	movlw	0x0A
	rcall	sendchar
	movlw	0x0D
	rcall	sendchar
	return



sendbhex:			;converts byte into ASCII and transmits
	movwf	temp4		;save byte
	swapf	temp4,W		;high nibble first		
	rcall	sendnib
	movf	temp4,W		;then low nibble
	rcall	sendnib
	movf	temp4,W		;put W back to initial value
	return



sendCRLF:			;send carriage return-line feed
	movlw	0x0A
	rcall	sendchar
	movlw	0x0D
	rcall	sendchar
	return



sendspce:
	movlw	' '
sendchar:			;bit-bang data out at 9600 baud
	movwf	temp3		;save data
	movlw	D'8'
	movwf	temp2		;bit counter
	bcf	SEROUT		;start bit is a low level
	call	wt104us		;wait one bit time
	nop
	nop			;trying to make all bits the same 
	nop
nxtbitsc:
	rrcf	temp3,F		;send lsb first
	bc	sendone
	nop
	bcf	SEROUT
	nop
	nop
drsc:	call	wt104us
	decfsz	temp2,F		;dec bit counter
	goto	nxtbitsc
	bsf	SEROUT		;stop bit is a high level
	call	wt104us		;leave output high between transmissions
	call	wt104us
	call	wt104us
	call	wt104us
	return

sendone:
	bsf	SEROUT		;if bit was 1, then high output
	bra	drsc


sendnib:			;converts low nibble into ASCII and transmits
	call	getASC		;just uses low nibble
	rcall	sendchar
	return

	endif

				;user settings byte (in flags2)
				;ints are already disabled when we do this
wrcfg:
	bcf	PIR1,EEIF	;clear write complete flag

	clrf	EEADR		;set up starting addr in EEPROM

	bcf	EECON1,EEPGD	;point to data memory
	bcf	EECON1,CFGS	;access EEPROM

	movf	flags2,W	;save settings
	movwf	EEDATA

	rcall	wrEEPROM

	incf	EEADR,F		;bump address
	comf	flags2,W
	movwf	EEDATA		;write complement to next address

	rcall	wrEEPROM

	return



wrEEPROM:
	bsf	EECON1,WREN 	;Enable writes

	movlw 	0x55		;Unlock write
	movwf 	EECON2
	movlw 	0xAA		;Interrupts should be disabled here
	movwf 	EECON2
	bsf 	EECON1,WR 	;Start the write
	nop
notfinwe:
	btfsc	EECON1,WR	;wait for write to complete
	bra	notfinwe

	bcf	EECON1,WREN 	;disable writes
	return



				;general-purpose delay routines
				;these use temp1, temp2, temp3
				;won't be very accurate with interrupts enabled

wt100ms	movlw	D'100'		;kill 100 mSec
	bra	skip22

wt50ms	movlw	D'50'		;kill 50 mSec
	bra	skip22

wt10ms	movlw	D'10'		;kill 10 mSec
	bra	skip22

wt8ms	movlw	D'8'		;kill 8 mSec
	bra	skip22

wt1ms	movlw	D'1'		;kill 1 mSec
skip22	movwf	temp2

wtolp	movlw	D'10'		;middle loop is 1.0 mSec
	movwf	temp3
	
;wtilp	movlw	D'53'		;1 cycle, delay for 100.15 uSec (26 MHz osc)
wtilp	movlw	D'215'		;1 cycle, delay for 100.15 uSec (26 MHz osc)
	movwf	temp1		;1 cycle	
dlyLp
	decfsz	temp1,F		;1 cycle, 2 if jump is taken
	bra	dlyLp		;2 cycles
				;inner loop total: ((215 * 3) + (1 * 4) + 2) * 153.85 nSec = 100.15 uSec
	decfsz	temp3,F	
	bra	wtilp		;inner loop until zero

	decfsz	temp2,F		;outer loop until zero
	bra	wtolp
	return

wt5sec	rcall	wt500ms
	rcall	wt500ms
	rcall	wt500ms
	rcall	wt500ms
wt3sec	rcall	wt500ms
	rcall	wt500ms
	rcall	wt500ms
	rcall	wt500ms
	rcall	wt500ms
	rcall	wt500ms
	return

wt500ms	rcall	wt100ms
wt400ms	rcall	wt100ms
	rcall	wt100ms
wt200ms	rcall	wt100ms
	rcall	wt100ms
	return


	ifndef	dbgser

wt104us:			;delay 104.2 uSec to get 9600 baud
	movlw	D'223'		;1 cycle, delay for 104 uSec (26 MHz clock)
	movwf	temp1		;1 cycle	
dlylpa:
	decfsz	temp1,F		;1 cycle, 2 if jump is taken
	bra	dlylpa		;2 cycles
				;1 cycle, total: ((224 * 3) + 4) = 676 cycles
				;using 223 due to ser code overhead
				;677 * 153.85 nSec = 104.15 uSec
	return			;2 + 2 for call

	endif


;; Above this line are our subroutines
;; *****************************************************************************
			
				;This is the actual start of the program.
start	clrwdt			;reset vector jumps here

	clrf	INTCON		;init intcon register, shut off all interrupts
	movlw	0x80
	movwf	INTCON2		;disable PORTB pullups
	clrf	PIE1		;init PIE1 register, shut off all interrupts
	clrf	PIE2

	clrf	PORTA		;select bank 0, FLIPTBE low

	movlw	0x08		;RDWDATb high, CLKRDAT low
	movwf	PORTB		;init port B

	movlw	0x60		;set EMPTY flag, serial data high
	movwf	PORTC

	movlw	PORTADIR	;set up I/O port directions
	movwf	TRISA

	movlw	PORTBDIR
	movwf	TRISB
	
	movlw	PORTCDIR
	movwf	TRISC

	movlw	0x0F
	movwf	ADCON1		;all A/D inputs are digital I/O
	movlw	0x07
	movwf	CMCON		;all comparators are off

	clrf	flags1		;clear flags

	btfsc	DRDY		;if DRDY is set, clear it
	bsf	CLKRDAT		;clock DRDY to toggle it
	nop
	bcf	CLKRDAT		;CLKRDAT sits low

	btfss	TBE		;if TBE not set, set it
	bsf	FLIPTBE
	nop
	bcf	FLIPTBE		;FLIPTBE sits low

;	movlw	0xA1
;	movwf	T1CON		;select 1:4 prescale for timer 1  (615.4 nSec/tick at 6.5/26 MHz)
				;set RD16, enable timer 1
				;16-bit timer will overflow after 26.2 mSec
				;that should be OK for even 300 baud
	rcall	wt100ms


;	call	rdcfg		;read settings from EEPROM (call when ints are disabled)
				;into flags 2

	movlw	0x07A
	movwf	T2CON		;select prescale = 16 for timer 2 (2.46 uSec/tick at 6.5/26 MHz)
				;select 1:16 postscale, timer 2 not enabled
	movlw	D'253'
	movwf	PR2		;select divide by 254 (624.84 uSec between overflows)
				;should be 10mSec between interrupts

	clrf	FSR1L		;rx buffer is empty
	clrf	FSR2L

	clrf	FIFOcnt		;no bytes in FIFO

	call	initser		;set up serial port

	ifdef	simtest

	call	myISR

	movlw	D'129'		;simulate writing 132 rec'd bytes to FIFO
	movwf	testctr
testwr:
	call	myISR
	decfsz	testctr,F
	bra	testwr

	call	myISR
	call	myISR
	call	myISR

	call	mainlp		;read a byte from FIFO

	movlw	D'129'		;simulate reading 132 bytes from FIFO
	movwf	testctr
testrd:
	call	mainlp
	decfsz	testctr,F
	bra	testrd

	call	mainlp
	call	mainlp
	call	mainlp

	endif

	clrf	PIR1		;clear interrupt flags
	clrf	PIR2
	clrf	RCON		;disable interrupt priority
	movlw	0x80
	movwf	INTCON2		;PORT B pullups diabled	
	clrf	INTCON3	
	movlw	0xC0
	movwf	INTCON		;turn on global interrupt enable bit
				;enable peripheral interrupts

	
mainlp:	
	movlw	0x02		;keep both FSR's pointed to buffer while we idle
	movwf	FSR1H		;pt into buffer
	movwf	FSR2H	

	btfsc	DRDY		;is rx data reg for CPU empty/available?
	bra	chk4rx

	btfss	DOUT		;did CPU read a byte we wrote earlier?
	bra	nodout
	bcf	DOUT		;yes, clear flag	
	decf	FIFOcnt,F	;and decrement FIFO count
				;we would not have set DOUT if count was 0
	bnz	dataout		;if count is 0 now
	bsf	EMPTY		;indicate FIFO is empty
	bra	chk4rx		;and skip to rx check
	
nodout:
				;rx data reg is empty, 
				;anything in the buffer?
	movf	FIFOcnt,W
	btfsc	STATUS,Z	;if so, send out the next byte
	bra	chk4rx		;otherwise skip to rx check

dataout:			;buffer is not empty, so send out another byte

	incf	FSR1L,F		;bump rd ptr, low byte only

	movf	PORTC,W
	andlw	0x30		;save current FIFO flags
	movwf	portcval

	movf	INDF1,W		;get byte from buffer
	andlw	0xF0		;keep high nibble
	iorlw	0x08		;keep RDWDATb high
	movwf	PORTB		;update PORT B with high nibble of byte

	movf	INDF1,W		;get byte from buffer
	andlw	0x0F		;keep low nibble
	iorwf	portcval,W
	movwf	PORTC		;update PORT C with low nibble of byte

	bsf	CLKRDAT
	nop
	bcf	CLKRDAT		;clk data into register, set DRDY flag
	
	bsf	DOUT		;and set flag indicating we wrote out a byte for CPU

	movlw	D'132'		;FIFO can still be full with a byte waiting for CPU to read
	subwf	FIFOcnt,W	;is FIFO still full?
	btfss	STATUS,Z
	bcf	FULL		;no, so clear flag

	bcf	EMPTY		;only clear FIFO empty flag when there is data for CPU to read
	
	bra	mainlp		;and loop


chk4rx:	
	btfss	NEWRX		;got a new byte?
	bra	chk4tx
	bcf	NEWRX		;clear flag
	btfsc	FULL
	bra	chk4tx		;if FIFO already full, just discard data

	incf	FSR2L,F		;bump wr ptr, low byte only

	movlw	D'131'
	subwf	FIFOcnt,W	;is FIFO almost full?
	btfsc	STATUS,Z
	bsf	FULL		;this byte will make it full, set flag

	movff	rxbyte,INDF2	;put MIDI rx byte in buffer
	incf	FIFOcnt,F	;and inc byte count


chk4tx:	btfsc	TBE		;any new data from the CPU?
	bra	mainlp		;no, so loop

	btfss	PIR1,TXIF	;is uart tx reg empty?
	bra	mainlp		;no, so loop

	movlw	0xF3
	movwf	TRISB		;set up port B data for input
	movlw	0xCF
	movwf	TRISC		;and port C as well

	bcf	RDWDATb		;enable tx data reg outputs
	movff	PORTB,portbval
	movff	PORTC,portcval	;read data
	bsf	RDWDATb		;float reg data outputs

	movlw	0x0F		;assemble byte we got
	andwf	portcval,F	;keep low 4 bits of C port
	movf	portbval,W
	andlw	0xF0		;and upper 4 of B port
	iorwf	portcval,W	;OR in C port bits

	movwf	txbyte		;save it

	xorlw	0xDD		;test for bank-change command
	bz	setbank

	movlw	0x03
	movwf	TRISB		;set up port B data for output
	movlw	0xC0
	movwf	TRISC		;and port C as well

	movf	txbyte,W
	movwf	TXREG		;write byte to PIC USART

	bsf	FLIPTBE		;toggle TBE to set it, to indicate that tx buffer is empty
	nop
	bcf	FLIPTBE

	bra	mainlp		;and loop


setbank:
	bsf	FLIPTBE		;toggle TBE to set it, to indicate that tx buffer is empty
	nop
	bcf	FLIPTBE
	
wt4bnk:	btfsc	TBE		;wait for another byte from CPU
	bra	wt4bnk

	bcf	RDWDATb		;enable tx data reg outputs
	movff	PORTB,portbval
	movff	PORTC,portcval	;read data
	bsf	RDWDATb		;float reg data outputs

	movlw	0x0F		;assemble byte we got
	andwf	portcval,F	;keep low 4 bits of C port
	movf	portbval,W
	andlw	0xF0		;and upper 4 of B port
	iorwf	portcval,W	;OR in C port bits

	andlw	0x1F		;just keep low 5 bits for bank, keep FLIPTBE low
	movwf	PORTA		;update bank bits

	movlw	0x03
	movwf	TRISB		;set up port B data for output
	movlw	0xC0
	movwf	TRISC		;and port C as well

	bsf	FLIPTBE		;toggle TBE to set it, to indicate that tx buffer is empty
	nop
	bcf	FLIPTBE

	bra	mainlp		;and loop
								
	END


		



	
	


